#!/usr/bin/env python

# Copyright and license: Please refer to LICENSE.TXT

import sys
import os
import os.path
import string
import re
import time
import random
import posixfile
import glob
import md5
import signal
import thread
import threading
from xml.dom import minidom
import htmlentitydefs
from xml.sax import saxutils

_selfpath=(os.path.dirname(os.path.realpath(__file__)).split('/'))[:-2]

from s2p_entities import *

# Paths definition
S2P_PATH_BASE		='/'.join(_selfpath)
CORE_PATH_BASE		= S2P_PATH_BASE+'/core'
APPS_PATH_BASE		= S2P_PATH_BASE+'/apps'

CORE_PATH_LIBEXEC	= CORE_PATH_BASE+'/libexec'
CORE_PATH_BIN		= CORE_PATH_BASE+'/bin'
CORE_PATH_CONFIG	= CORE_PATH_BASE+'/config'
CORE_PATH_VH		= CORE_PATH_BASE+'/vh'
CORE_PATH_FVH		= CORE_PATH_BASE+'/fvh'
CORE_PATH_MISC		= CORE_PATH_BASE+'/misc'
CORE_PATH_SDK		= CORE_PATH_BASE+'/sdk'

S2P_XML_GARBAGE_LOCK	= threading.Lock()
S2P_XML_GARBAGE_STARTED	= False

S2P_XML_CACHE={}
S2P_XML_CACHE[0]={}
S2P_XML_CACHE[1]={}

S2P_CONF_SYSTEM	= {}

debug=0

def open_anything(source):            
	if hasattr(source, "read"):
		return source

	if source == '-':
		import sys
		return sys.stdin

	import urllib
	try:
		return urllib.urlopen(source)
	
	except (IOError, OSError):
		pass

	try:
		return open(source)
	except (IOError, OSError):
		pass

	import StringIO
	return StringIO.StringIO(str(source))

def write_file(file, text):
	fp = posixfile.open(file, 'w')
	fp.lock('w|')
	fp.write(text)
	fp.lock('u')
	fp.close()

def _xml_to_array2(node):
	
	out={}
	out['+']={} # Childs dictionary
	out['#']={} # Values dictionary
	out['@']={} # Attributes dictionary
	
	if node.attributes:
		for key in node.attributes.keys():
			out['@'][str(key)]=node.attributes[str(key)].value
	
	childs=node.childNodes
	for child in childs:
		
		if child.nodeType==node.TEXT_NODE:
			out['#']=child.data
			out['#']=re.sub(r'^[\n\r\s]+', '', out['#'])
			out['#']=re.sub(r'[\n\r\s]+$', '', out['#'])
			
		elif child.nodeType==node.ELEMENT_NODE:
			n=child.tagName
			if not out['+'].has_key(n):
				out['+'][n]={}

			out['+'][n][len(out['+'][n].keys())]=_xml_to_array2(child)
	
	return out

def xml_simplify_array(a):
	if not a.has_key('+'): return a

	if a.has_key('@') and not len(a['@']): del a['@']
	if a.has_key('#') and not len(a['#']): del a['#']

	if a.has_key('#') and len(a['#']): return a['#']

	for k in a['+'].keys():

		if len(a['+'][k]) == 1:
			a[k]=xml_simplify_array(a['+'][k][0])
		else:
			for n in a['+'][k].keys():
				if not a.has_key(k): a[k]={}
				
				a[k][n]=xml_simplify_array(a['+'][k][n])

	del a['+']

	return a

def s2p_xml_garbage_collector():
	global S2P_XML_CACHE
	global S2P_XML_GARBAGE_LOCK
	
	try:
		timeout=int(S2P_CONF_SYSTEM['tuning']['xml-cache']['timeout'])
		
	except:
		timeout=60
		
	while True:
		time_now=time.time()
		
		S2P_XML_GARBAGE_LOCK.acquire()
		
		for k in S2P_XML_CACHE[0].keys():
			if time_now-S2P_XML_CACHE[0][k][1]>timeout:
				del S2P_XML_CACHE[0][k]
				
		for k in S2P_XML_CACHE[1].keys():
			if time_now-S2P_XML_CACHE[1][k][1]>timeout:
				del S2P_XML_CACHE[1][k]
				
		S2P_XML_GARBAGE_LOCK.release()
		
		time.sleep(1)

def array_to_xml2_simple(array, depth=0):
	out=[]

	indent=""
	for i in range(0, depth): indent=indent+"\t"
	
	for k in array.keys():
		if not array[k]:
			out.append(indent+"<"+k+"/>")
		elif type(array[k]) == dict:
			out.append(indent+"<"+k+">")
			out.append(array_to_xml2_simple(array[k], depth+1))
			out.append(indent+"</"+k+">")
		else:
			out.append(indent+"<"+k+">"+s2p_xml_escape(array[k])+"</"+k+">")

	return "\n".join(out)

def array_to_xml2(array, depth=0):
	out=[]

	indent=""
	for i in range(0, depth): indent=indent+"\t"
	
	for k in array.keys():
		for v in array[k].values():

			args_a=[]
			if len(v['@']):
				for w in v['@'].keys():
					args_a[w+"=\""+s2p_xml_escape(re.sub(r'"', '&quot;', v['@'][k]))+"\""]

			args=""
			if len(args_a): args=" ".join(args_a)

			if not len(v['+']) and len(v['#']):
				out.append(indent+"<"+k+args+">"+s2p_xml_escape(v['#'])+"</"+k+">")
			
			elif not len(v['+']) and not len(v['#']):
				out.append(indent+"<"+k+args+"/>")
				
			else:
				out.append(indent+"<"+k+args+">")
				out.append(array_to_xml2(v['+'], depth+1))
				out.append(indent+"</"+k+">")
	
	return "\n".join(out)

def array_to_xml(array, root="config", simple=True):
	out=[]
	out.append("<?xml version=\"1.0\" encoding=\"utf-8\"?>")
	out.append("<"+root+">")

	if simple:
		out.append(array_to_xml2_simple(array, 1))
	else:
		out.append(array_to_xml2(array, 1))
	
	out.append("</"+root+">")

	return "\n".join(out)

def xml_to_array(filename, simple=1, cache=False):
	global S2P_XML_CACHE
	global S2P_CONF_SYSTEM
	
	global S2P_XML_GARBAGE_LOCK
	
	if simple:	simple=1
	else:		simple=0
	
	if cache:
		md5cache_key=md5sum(filename)
		
		try:
			if S2P_CONF_SYSTEM['tuning']['xml-cache']['enabled'] != 'true': cache=False
		except:
			cache=False
	
	S2P_XML_GARBAGE_LOCK.acquire()
	
	if cache and S2P_XML_CACHE[simple].has_key(md5cache_key):
		S2P_XML_CACHE[simple][md5cache_key][1]=time.time()
		out=S2P_XML_CACHE[simple][md5cache_key][0]
		S2P_XML_GARBAGE_LOCK.release()
		return out
		
	S2P_XML_GARBAGE_LOCK.release()
	
	f = open_anything(filename)
	xml_text = f.read()
	f.close()

	res=re.findall(r'<\?include ([^ ]+) \?>', xml_text)
	if res:
		for match in res:
			include_file=os.path.dirname(filename)+"/"+match

			if not os.path.isfile(os.path.dirname(filename)+"/"+match): continue
		
			f=open(include_file, "r")
			xml_text_include = f.read()
			f.close()
			
			xml_text=re.sub(r'<\?include '+match+' \?>', xml_text_include, xml_text)

	xml_text=htmlunescape(xml_text)

	if not simple:
		dom=minidom.parse(open_anything(xml_text))
		out=_xml_to_array2(dom.documentElement)
		
		if cache:
			S2P_XML_GARBAGE_LOCK.acquire()
			S2P_XML_CACHE[simple][md5cache_key]=[out['+'], time.time()]
			out=S2P_XML_CACHE[simple][md5cache_key][0]
			S2P_XML_GARBAGE_LOCK.release()
			return out

		return out['+']

	else:
		dom=minidom.parse(open_anything(xml_text))
		out=_xml_to_array2(dom.documentElement)
		
		if cache:
			S2P_XML_GARBAGE_LOCK.acquire()
			S2P_XML_CACHE[simple][md5cache_key]=[xml_simplify_array(out), time.time()]
			out=S2P_XML_CACHE[simple][md5cache_key][0]
			S2P_XML_GARBAGE_LOCK.release()
			return out

		return xml_simplify_array(out)

def systempt(cmd):
	try:
		f=os.popen(cmd)
		out=f.read()
		f.close()
	except:
		return ""
	
	return re.sub(r'[\n\r]+$', '', out)

def debugprint(message):
	if debug: print message

def md5sum(t):
	m = md5.new()
	m.update(t)
	return m.hexdigest()

def random_string():
	letters = "abcdefghijklmnopqrstuvwxyz"
	letters += letters.upper () + "0123456789"
	
	random_string = ''
	for i in range(16):
		random_string += random.choice(letters)

	return random_string

def ini_to_array(filename):
	out={}
	f=open(filename, 'r')
	
	for line in f:
		if line == "": continue
		if re.match(r'^[\#\;\[]', line): continue
		line=re.sub(r'[\n\r]+', '', line)
		
		couple=line.split('=', 2)
		out[couple[0]]=couple[1]
		
	f.close()
			
	return out

def read_file(filename):
	if not os.path.isfile(filename): return ""
	
	f=open(filename, 'r')
	out=""
	for line in f: out=out+line
	f.close()
			
	return out

def fix_digits(number, digits):
	out=str(number)

	while len(out) < digits:
		out="0"+out

	return out

def trim(s):
	s=re.sub(r'^[\n\r\s]+', '', s)
	s=re.sub(r'[\s\n\r]+$', '', s)
	return s
			
def sprintf(format, *args):
	return format % args,

def set_vars(txt, replace):
	if not txt: txt=""
	
	for k in replace.keys():

		if not replace[k]: replace[k]=""
		txt=re.sub(r'\$\{'+k+'\}', replace[k], txt)

	return txt

def string_escaped_to_octal(txt):
	out=""
	pos=0

	if not len(txt): return ""

	while pos<len(txt):
		c=txt[pos]

		if c == '\\':
			pos+=1
			c=txt[pos]

			if c == 'b': out+='\\010'
			elif c == '"': out+='\\042'
			elif c == '\\': out+='\\134'
			elif c == 'f': out+='\\014'
			elif c == 'n': out+='\\012'
			elif c == 'r': out+='\\015'
			elif c == 't': out+='\\011'

			pos+=1
			
		else:
			out+=c
			pos+=1

	return out
			
def escape_string(txt):
	out=""
	pos=0
	
	if not len(txt): return ""

	while pos<len(txt):
		c=txt[pos]

		if c == '\b': out+='\\b'
		elif c == '"':  out+='\\"'
		elif c == '\\': out+='\\\\'
		elif c == '\f': out+='\\f'
		elif c == '\n': out+='\\n'
		elif c == '\r': out+='\\r'
		elif c == '\t': out+='\\t'
		elif ord(c) < ord(' ') or ord(c) > 0177:
			out+='\\'
			out+=str((ord(c) >> 6) & 07)
			out+=str((ord(c) >> 3) & 07)
			out+=str(ord(c) & 07)
			
		else:
			out+=c

		pos+=1

	return out

def unescape_string(txt):
	out=""
	pos=0
	
	if not len(txt): return ""

	while pos<len(txt):
		c=txt[pos]

		if c == '\\':
			pos+=1
			c=txt[pos]

			if 	c == '0' or c == '1' or c == '2' or c == '3' or \
				c == '4' or c == '5' or c == '6' or c == '7' \
			:
				octal=0
				init_pos=pos

				while pos < init_pos + 3 and (
					c == '0' or c == '1' or c == '2' or c == '3' or \
					c == '4' or c == '5' or c == '6' or c == '7' \
				):
					octal+=int(txt[pos])*(8 ** (init_pos-pos+2))
					pos+=1

				out+=chr(octal)

			elif c == 'b':
				out+='\b'
				pos+=1

			elif c == 'f':
				out+='\f'
				pos+=1

			elif c == 'n':
				out+='\n'
				pos+=1

			elif c == 'r':
				out+='\r'
				pos+=1

			elif c == 't':
				out+='\t'
				pos+=1

			else:
				out+=c
				pos+=1
	
		else:
			out+=c
			pos+=1

	return out

def get_timestamp():
	return time.strftime("%Y%m%d%H%M%S", time.localtime())

def in_array(haystack, needle):
	for i in haystack:
		if i == needle: return True

	return False

def get_registered_pid(s):
	global CORE_PATH_CONFIG
	S2P_CONF_SYSTEM	= xml_to_array(CORE_PATH_CONFIG+'/system.xml')
	
	try:
		fp=open(S2P_CONF_SYSTEM['path']['run']+'/'+s+".pid", "r")
		out=re.sub(r'[ \n\r]+', '', fp.read())
		fp.close()

		return out
	
	except:
		return 0

def s2p_register_pid(s, lock=True, pid=None):
	global CORE_PATH_CONFIG
	S2P_CONF_SYSTEM	= xml_to_array(CORE_PATH_CONFIG+'/system.xml')

	if not pid: pid=os.getpid()
	
	if lock and os.path.isfile(S2P_CONF_SYSTEM['path']['run']+'/'+s+".pid"):
		print "Another instance is already running"
		print S2P_CONF_SYSTEM['path']['run']+'/'+s+".pid exists"
		sys.exit(-1)

	fp=open(S2P_CONF_SYSTEM['path']['run']+'/'+s+".pid", "w")
	fp.write(str(pid))
	fp.close()

	return pid

def s2p_deregister_pid(s):
	global S2P_CONF_SYSTEM

	S2P_CONF_SYSTEM	= xml_to_array(CORE_PATH_CONFIG+'/system.xml')
	try:
		os.unlink(S2P_CONF_SYSTEM['path']['run']+'/'+s+".pid")
	except:
		1

def s2p_xml_escape(text):
	return htmlescape(text)

# System configuration
S2P_CONF_SYSTEM	= xml_to_array(CORE_PATH_CONFIG+'/system.xml')

try:
	if S2P_CONF_SYSTEM['tuning']['xml-cache']['enabled'] == 'true':
		
		S2P_XML_GARBAGE_LOCK.acquire()
		if not S2P_XML_GARBAGE_STARTED:
			thread.start_new_thread(s2p_xml_garbage_collector, ())
			S2P_XML_GARBAGE_STARTED=True
		S2P_XML_GARBAGE_LOCK.release()
		
		# Used for threading
		signal.signal(signal.SIGINT, signal.SIG_DFL)
		
except:
	1
